open Printf

type coord = {
   x: int;
   y: int
}

let move c x_off y_off =
   let newc = {x = c.x + x_off; y = c.y + y_off } in
      newc;;

let distance a b = 
   abs(a.x - b.x) + abs(a.y - b.y);;
   
type package = {
   mutable pos: coord;
   id: int;
   weight: int;
   dst: coord
}

type direction = 
     North
   | South
   | East
   | West;;


let direction_of_char c = 
   match c with 
       'N' -> North
     | 'S' -> South
     | 'E' -> East
     | 'W' -> West
     | _ -> raise (Failure ("Bad direction : " ^ Char.escaped c))

let direction_of_string s = 
   match s with 
       "N" -> North
     | "S" -> South
     | "E" -> East
     | "W" -> West
     | _ -> raise (Failure ("Bad direction : " ^ s))

let char_of_direction dir = 
   match dir with
       North -> 'N'
     | South -> 'S'
     | East -> 'E'
     | West -> 'W'

let move_dir c (dir:direction) = 
   match dir with
       North -> { x = c.x; y = c.y + 1 }
     | South -> { x = c.x; y = c.y - 1 }
     | East -> { x = c.x + 1; y = c.y }
     | West -> { x = c.x - 1; y = c.y }

let by_distance c = fun (x1,y1) (x2,y2) ->
   (distance c y1.dst) - (distance c y2.dst);;

let direction_of_move src dst =
   if src.x < dst.x then
      East	 
   else (
      if src.x > dst.x then
	 West
      else (
	 if src.y > dst.y then South
	 else North
      )
   )
	 
	 
	 


class ['a] interval first last = object(self)
   val mutable fst = (first:'a)
   val mutable lst = (last:'a)

   method fst = fst
   method lst = lst
		   
   (* coord inclusion *)
   method incl x = x >= fst && x <= lst

   (* interval intersection *)
   method int_inter (other:'a interval) =
      (fst >= other#fst && fst <= other#lst) || (fst <= other#fst && lst >= other#fst)
      
   method extend x = 
      if fst = -1 then    
	 fst <- x
      else 
	 ();
      lst <- x
	 
   method to_string =
      sprintf "(%i,%i)" fst lst;

end;;

class zone (width:int interval) (height:int interval) = object(self)
   val w = width
   val h = height
	      
   val mutable relatives = ([]:(zone ref * direction) list)

   method w = w
   method h = h
		 
   (* inclusion *)
   method incl point = 
      w#incl point.x && h#incl point.y
	 
   method relatives = relatives

   method add_relative z = 
      let ori = 
	 if w#fst > !z#w#lst then
	    West
	 else if w#lst < !z#w#fst then
	    East
	 else if h#fst > !z#h#lst then
	    South
	 else 
	    North
      in  
	 relatives <- (z, ori) :: relatives
	    
   method to_string = 
      let s = sprintf "Zone: (%i,%i) -> (%i,%i) " w#fst h#fst w#lst h#lst in
	 s ^ String.concat " " (List.map (fun (x, dir) -> sprintf "%c [(%i,%i);(%i,%i)]" 
					     (char_of_direction dir) !x#w#fst !x#h#fst !x#w#lst !x#h#lst)
				   relatives)
   method orientation (z:zone ref) = 
      let tmp = List.find (fun (x,y) -> !x = !z) relatives in
	 snd tmp;
   
   method intersect_horiz inter = 
      w#int_inter inter;
      
   method intersect_vert inter = 
      h#int_inter inter;
      
   method intersect (z:zone) = 
      w#int_inter z#w && h#int_inter z#h;
      
end;;

exception Found
exception Cycle
exception No_path

class path s (src:coord) (dst:coord) = object(self)
   val mutable p = (s:(zone ref) Stack.t)
   val src = src
   val dst = dst
		
   method dst = dst
   method src = src

   method length = 
      (*false*)
      Stack.length p
	 
   method contains v =
      try 
	 Stack.iter (fun x -> if x = v then raise Found) p;
	 false
      with 
	  Found ->
	     true
	     
   method empty =
      Stack.is_empty p;

   method top =
      Stack.top p;

   method pop = 
      Stack.pop p;

   method push v = 
      Stack.push v p;

   method copy = 
      new path (Stack.copy p) src dst
	 
   method rev = 
      let tmp = Stack.create () in
	 Stack.iter (fun x -> Stack.push x tmp) p;
	 p <- tmp;	 

end;;

let create_path src dst = 
   new path (Stack.create ()) src dst;;
      
let zone_includep p =
   fun x -> x#incl p;;

let find_shortest (pathes:(path option) list) = 
   match pathes with
       [] -> raise Not_found
     | _ ->
	  List.hd (List.sort (fun x y -> 
				 match x with
				     Some a ->
					(match y with 
					     Some b ->
						a#length - b#length
					   | None ->
						-1;
					) 
				   | None -> 1;)
		      pathes);
	  
